/*
 *  Paul Regenhardt  10/13/01
 *  This program reads the Old MS DOS control.nnn and backup.nnn
 *  programs and restores them on a new system.  I was worried
 *  about using the MS Restore on Windows as I converted to FAT
 *  32 a while ago.
 ***************************************************************
 *
 * The backup.nnn file contains the actual contents of all the
 * data in a nice unencrypted format.
 * The control.nnn file contains the file name and offsets of the 
 * files.  The record structure of the control file is:
 *
 *	Offset	DataType	Description
 *  0		byte		offset to the first directory information
 *  1-6		String		The string "BACKUP"
 *  7		byte		The control file
 *  n-1     byte        0 - not the end, 0xFF End of disks
 *  Note that there is no version number in the file, so it's 
 *   difficult to know what backup version you got.
 ***************************************************************
 *
 * Directory structure
 *
 *	Offset	DataType	Description
 *  n		Char		Drive Letter of the original copy
 *  n+1		String		Null terminated directory name
 *  n+0x40  6 bytes     Unknown, probably Timestamp, directory 
 *						attributes
 *******************************************************************
 *
 * File structure
 *
 *	Offset	DataType	Description
 *  0		byte		Should be 34, length of record
 *  1		String		12 character file name (includes dot and 
 *						extension)
 *  13      byte		A "3", not sure what this is, file attributes?
 *  14      I*4		    Total File Length
 *  18      I*2			1 or 2 in my tests, seems to be the  
 *						diskette number it spanned across?
 *  20		I*4			Offset in the backup.nnn file to the start 
 *						of this file
 *  24		I*4			Number of bytes in the backup.nnn file to 
 *						copy out, can be less than the size when 
 *						the file spans multiple diskettes
 *  28      ?			I suspect the next six bytes are a Timestamp
 *******************************************************************
 *
 * Without knowing all the details this parser does what it can.
 * the big problem is that when you run into a new directory it
 * is difficult to tell from a file.  I'm assuming if the next record
 * starts and isn't a 0x22, it's a directory entry.
 * A lot of this is guess work as I don't have a lot of file, 
 * hopefully it will work for you.
 *
 ***************************************************************
 * Program Arguments
 * 
 * Originating disk:\directory where control file lives
 * Destination disk:\directory destination root
 */
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <direct.h>
#include <stdlib.h>

char directory[200];
char destination[200];

void createDirectory(char *directory)
{
	char *d;
	char subdir[200];
	FILE* file;

	d = strchr(directory, '\\');
	while (d != NULL) {
		strcpy(subdir, destination);
		strcat(subdir, "\\");
		strncat(subdir, directory, d-directory);
		strcat(subdir, "\0");
		if ((file = fopen(subdir, "r")) != NULL) {
			fclose(file);
		} else {
			_mkdir(subdir);
		}
		d = strchr(d+1, '\\');
	}

	strcpy(subdir, destination);
	strcat(subdir, "\\");
	strcat(subdir, directory);

	if ((file = fopen(subdir, "r")) != NULL) {
		fclose(file);
	} else {
		_mkdir(subdir);
	}
}


void processFile(FILE* backup, char* fName, unsigned int fLen, 
				 fpos_t fPntr, unsigned int nBytes, short append)
{
	FILE* file;
	unsigned int bSize = 32000;
	unsigned int bRead, n;
	char* buffer;
	char name[128];

	strcpy(name, destination);
	strcat(name, "\\");
	strcat(name, directory);
	strcat(name, "\\");
	strcat(name, fName);

	fsetpos(backup, &fPntr);
	if (append == 1)
		file = fopen(name, "w");
	else // append
		file = fopen(name, "a");

	if (file == NULL) {
		printf("\nUnable to write %s", name);
		return;
	}

	printf("\nWriting file %s", name);

	// copy the bytes
	n = nBytes;
	if (nBytes < bSize) bSize = nBytes;
	buffer = malloc(bSize);

	while (n > 0) {
		if (n < bSize) bSize = n;
		bRead = fread(buffer,sizeof( char ),bSize, backup);
		fwrite(buffer, sizeof( char ), bRead, file);
		n -= bSize;
	}

	fclose(file);
	free(buffer);
}


// return 0 as long as their is another control file to parse
int parseFiles(FILE* control, FILE* backup)
{
	char fName[13];
	fpos_t fPntr, bOffset;
	unsigned int  fLen, nBytes, off;
	unsigned short dNum;
	unsigned char b, lastDisk, lastDiskette;
	int val=0;

	lastDiskette = fgetc(control);  // if zero, there are more diskettes
	fgetpos(control, &fPntr);
	b = fgetc(control);

	while (!feof(control)) {// return when the end of the control file is reached.
		if (b == 34) {	// if 34 its a file
			fgets(fName, 12, control);
			bOffset = fPntr + 14;
			fsetpos(control, &bOffset); // skipping the "3" byte, whatever that is
			fread((char*) &fLen, sizeof (unsigned int), 1, control); /* get the length by reading 4 raw bytes into an I*4 */
			fread((char*) &dNum, sizeof (unsigned short), 1, control); // load disk#
			fread((char*) &off, sizeof (unsigned int), 1, control); // load the offset in the backup.nnn file for this file
			bOffset = off;
			fread((char*) &nBytes, sizeof (unsigned int), 1, control);  // load the number of bytes in this backup.nnn to read
			processFile(backup, fName, fLen, bOffset, nBytes, dNum);
			fPntr += 34;
		} else { // otherwise it's a directory
			lastDisk = b;      // drive letter
			fscanf(control,"%s", directory);  // directory now contains the directory name
			// Create the directory if it doesn't exist
			createDirectory(directory);
			fPntr += 70;
		}
		val = fsetpos(control, &fPntr);
  	    b = fgetc(control);
	}
	return lastDiskette;
}


main(int argc, char**argv)
{
	char *source;
	int  controlNumber=1;
	FILE *control, *backup;
	char cExt[4];
	char fName[200];
	char c;
	fpos_t fPntr;
	int  cNum;
	char buffer[100];
	unsigned char b, offset;
	int status = 0;

	if (argc < 2) { // print out how to use the tool
		printf("\nUsage:  restore source destination\n");
		return(1);
	}

	source = argv[1];

	if (argc == 3) { // set the current working directory to the destination
		strcpy(destination, argv[2]);
	} else {
		strcpy(destination, ".");
	}

	while( status == 0 ) {
		// open up the first control file
		sprintf(cExt,"%03d", controlNumber);
	
		strcpy(fName, source);
		strcat(fName, "\\");
		strcat(fName, "control.");
		strcat(fName, cExt);
		control = NULL;
		while (control == NULL) {
			control = fopen(fName, "r");
			if (control == NULL) {
				printf("\nInsert diskette %d and hit RETURN, or ^C to abort...", controlNumber);
				scanf("%c", &c);
				if (c != 10) return (1);
			} else {
				/* read the header */
				offset = fgetc(control);
				fscanf(control,"%s", &buffer);
				/* Is is a valid backup file? */
				if (strncmp(buffer,"BACKUP", 6) != 0) {
					printf("\nSorry, this is not a compatible backup file");
					return(1);
				}
				fgetc(control);
				fgetc(control);
				b = fgetc(control);
				cNum = b;
				/* Got the right diskette loaded? */
				if (cNum != controlNumber) {
					fclose(control);
					control = NULL;
					printf("\nDisk %d inserted by mistake.  Please insert diskette %d and hit RETURN, or ^C to abort...", cNum, controlNumber);
					scanf("%c", &c);
					if (c != 10) return (1);
				}
			}
		}

		/* Open the backup file */
		strcpy(fName, source);
		strcat(fName, "\\");
		strcat(fName, "backup.");
		strcat(fName, cExt);
		backup = fopen(fName, "r");
		if (backup == NULL) {
			printf("\nError opening %s, operation aborts", fName);
			return(1);
		}

		fPntr = offset-1;
		fsetpos(control, &fPntr);

		/* Start Parsing! */
		status = parseFiles(control, backup);
		controlNumber++;
		fclose(backup);
		fclose(control);
	}
}

